package automaton;
import heap.*;
import code.*;
import java.util.*;


/** Represents a link between an initial and a final {@link State} in an {@link Automaton}. 
 *  This Transition occurs when his guard is true, and his actions are triggered.
 *  @author Swann Perarnau 
 * 	@version 1.0
 */
public class Transition {
	/** Holds the initial state. */
	private State sinitial;
	
	/** Holds the final state. */
	private State sfinal;
	
	/** Holds the guard expression as a {@link CodeNode} tree. */
	private CodeNode guard;
	
	/** Holds the actions statements as a {@link CodeNode} tree. */
	private CodeNode actions;
	
	/** Creates a transition between sinitial and sfinal with following guard and actions. 
	 * @param sinitial	The initial state of this transition.
	 * @param sfinal	The final state of this transition.
	 * @param guard		A representation of this transition's guard.
	 * @param actions	A representation of this transition's actions.
	 * 
	 * */
	public Transition(State sinitial, State sfinal, CodeNode guard, CodeNode actions) {
		this.sinitial = sinitial;
		this.sfinal = sfinal;
		this.guard = guard;
		this.actions = actions;
	}
	
	/** Creates a void transition */
	public Transition()
	{
		this.sinitial = null;
		this.sfinal = null;
		this.guard = null;
		this.actions = null;
	}
	
	
	public CodeNode getActions() {
		return actions;
	}

	public void setActions(CodeNode actions) {
		this.actions = actions;
	}

	public CodeNode getGuard() {
		return guard;
	}

	public void setGuard(CodeNode guard) {
		this.guard = guard;
	}

	public State getSfinal() {
		return sfinal;
	}

	public void setSfinal(State sfinal) {
		this.sfinal = sfinal;
	}

	public State getSinitial() {
		return sinitial;
	}

	public void setSinitial(State sinitial) {
		this.sinitial = sinitial;
	}
	
	public void heq(Map m)
	{
	    /* Swann's code: generates parallel assignments
		Set e = m.entrySet();
		Iterator i = e.iterator();
		while(i.hasNext())
		{
			Map.Entry tmp = (Map.Entry) i.next();
			this.actions = CodeNode.slist(this.actions, CodeNode.actions_xy(((Vertex)tmp.getKey()).counter(), ((Vertex)tmp.getValue()).counter()));
		}
	    */




	    /* Marius code, generate sequential assignments */

	    /* 1st step, build the counters map String -> String */
	    HashMap counterMap = new HashMap();  
	    Iterator i = m.entrySet().iterator();
	    while (i.hasNext()) {
		Map.Entry tmp = (Map.Entry) i.next();
		String key = ((Vertex)tmp.getKey()).counter();
		String value = ((Vertex)tmp.getValue()).counter();
		if ( !key.equals(value)) {
		    counterMap.put( key, value );
		    // debug: System.out.print(key + " := " + value + " | ");
		}
	    }

	    /* 2nd step, generate assignments */
	    if (!counterMap.isEmpty()) {
		// debug: System.out.print("\t-->\t");
	    
		while ( !counterMap.isEmpty() ) {
		    // go through the map, if there is a (key, value)
		    // pair such that the key is not present as value
		    // in the map, write down the assignment, and
		    // continue else, we introduce one tmp variable,
		    // rewrite assignements and continue
		    Iterator it = counterMap.keySet().iterator();
		    String key = null;
		    while ( it.hasNext() ) {
			String k = (String) it.next();
			if ( !counterMap.containsValue(k) ) {
			    key = k; break;
			}
		    }
		    // depending if a good key has been found
		    if (key != null) {
			// write down the assignment, reduce the map
			// and continue
			String value = (String) counterMap.get( key );
			this.actions = CodeNode.slist( this.actions, CodeNode.actions_xy(key, value));
			// debug: System.out.print(key + " := " + value + " ; ");
			counterMap.remove( key );
		    } 
		    else {
			// there are circular assignments, break them
			// using a 'tmp' variable remember the first
			// assigned key, then rewrrite all its right
			// occurrences using 'tmp'
			Iterator jt = counterMap.keySet().iterator();
			while ( jt.hasNext() ) {
			    String k = (String) jt.next();
			    if ( key == null ) {
				// this is the first assignment, store the value of key into 'tmp'
				key = k;
				this.actions = CodeNode.slist( this.actions, CodeNode.actions_xy("tmp", key));
				// debug: System.out.print("tmp := " + key +" ; ");
			    }
			    String value = (String) counterMap.get( k );
			    if ( value.equals(key) )
				counterMap.put( k, "tmp" );
			}
		    }
		    
		}  
		// debug: System.out.println(""); 
	    }
	    
	}

	public int hashCode() {
		final int PRIME = 31;
		int result = 1;
		result = PRIME * result + ((actions == null) ? 0 : actions.hashCode());
		result = PRIME * result + ((guard == null) ? 0 : guard.hashCode());
		return result;
	}

	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		final Transition other = (Transition) obj;
		if(this.sfinal != other.sfinal)
			return false;
		if(this.sinitial != other.sinitial)
			return false;
		if (actions == null) {
			if (other.actions != null)
				return false;
		} else if (!actions.equals(other.actions))
			return false;
		if (guard == null) {
			if (other.guard != null)
				return false;
		} else if (!guard.equals(other.guard))
			return false;
		return true;
	}
	
	
}
