package code;
import java.util.*;
import code.parser.*;
import output.*;


/** Represents a node of the abstract syntax tree of a program.
 *  This class implements methods used by the parser and a lot of "macros"
 *  for building expressions and statements.
 *  
 *  NOTA : this class may be modified deeply as this project is written
 *  (some methods will probably move to subclasses)
 * 
 * @author Swann Perarnau
 * @version 1.2
 *  */

public abstract class CodeNode implements Node {
	
	/** The parent Node in the AST of this Node. */
	protected Node parent;
	
	/** All the children in the AST of this Node. */
	protected Node[] children;
	
	/** The type of this Node. This type is termined when building the AST as the program source is parsed.
	 *  @see code.parser.CodeParserTreeConstants
	 * */
	protected int id;
	
	/** The CodeParser which generated this Node and all the AST.
	 * @see code.parser.CodeParser
	 * */
	protected CodeParser parser;

	
	/**The number of an instruction in this program.*/
	private int number;

	/** Type of Node : this Node must be typechecked.*/
	public static final int UNCHECKED = -1;
	
	/** Type of Node : this Node is a statement in the program.
	 * @see code.Statement
	 * @see code.CNIf
	 * @see code.CNAsgn
	 * @see code.CNWhile
	 * @see code.CNSList
	 * @see code.CNProgram
	 * */
	public static final int STATEMENT =  0;
	
	/** Type of Node : this Node is part of an integer expression.
	 * @see code.CNAdd
	 * @see code.CNSub
	 * @see code.CNMult
	 * @see code.CNId
	 * @see code.CNInt
	 * */
	public static final int INTEGER   =  1;
	
	/** Type of Node : this node is part of a pointer expression.
	 * @see code.CNNew
	 * @see code.CNId
	 * @see code.CNNull
	 * */
	public static final int POINTER   =  2;
	
	/** Type of Node : this node is part of a boolean expression.
	 * @see code.CNNot
	 * @see code.CNAnd
	 * @see code.CNEq
	 * @see code.CNGe
	 * @see code.CNGt
	 * @see code.CNLe
	 * @see code.CNLt
	 * @see code.CNOr
	 * */
	public static final int BOOLEAN   =  3;

	
	/** Represents the type of this Node is the AST */
	protected int type = UNCHECKED;

	public CodeNode(int i)
	{
		id = i;
	}

	public CodeNode(CodeParser p,int i)
	{
		this(i);
		parser = p;
	}

	public void jjtOpen() {
	}

	public void jjtClose() {
	}

	public void jjtSetParent(Node n) { parent = n; }
	public Node jjtGetParent() { return parent; }

	public void jjtAddChild(Node n, int i) {
		if (children == null) {
			children = new Node[i + 1];
		} else if (i >= children.length) {
			Node c[] = new Node[i + 1];
			System.arraycopy(children, 0, c, 0, children.length);
			children = c;
		}
		children[i] = n;
	}

	public Node jjtGetChild(int i) {
		return children[i];
	}

	public int jjtGetNumChildren() {
		return (children == null) ? 0 : children.length;
	}

	public String toString() { return CodeParserTreeConstants.jjtNodeName[id]; }
	public String toString(String prefix) { return prefix + toString(); }

	public void dump(String prefix) {
		System.out.println(toString(prefix));
		if (children != null) {
			for (int i = 0; i < children.length; ++i) {
				CodeNode n = (CodeNode)children[i];
				if (n != null) {
					n.dump(prefix + " ");
				}
			}
		}
	}


	public abstract int typeCheck(Map types);

	public int type()
	{
		return type;
	}
	
	public void setNumber(int n)
	{
		number = n;
	}
	
	public int getNumber()
	{
		return number;
	}

	public abstract String out(FileOutput module);


	public boolean equals(Object o)
	{
		if(this == o)
			return true;
		if(o == null)
			return false;
		if(this.getClass() != o.getClass())
			return false;
		CodeNode other = (CodeNode)o;
		if(other.type != this.type)
			return false;
		if(other.jjtGetNumChildren() != this.jjtGetNumChildren())
			return false;

		for(int i=0;i<this.jjtGetNumChildren();i++)
		{
			if(!this.jjtGetChild(i).equals(other.jjtGetChild(i)))
				return false;
		}
		return true;
	}
        
        public boolean isVoid() {
	        return false;
        }

	public static CodeNode voidGuard()
	{
		return null;
	}

	public static CodeNode voidActions()
	{
		return null;
	}	

	public static CodeNode and(CodeNode a, CodeNode b)
	{
		if(a == null && b == null)
			return null;
		if(a == null && b!= null)
			return b;
		if(a != null && b == null)
			return a;		
		CNAnd n = new CNAnd(code.parser.CodeParserTreeConstants.JJTAND);
		n.jjtAddChild(a, 0);
		n.jjtAddChild(b, 1);
		return n;
	}

	public static CodeNode not(CodeNode a)
	{
		CNNot n = new CNNot(code.parser.CodeParserTreeConstants.JJTNOT);
		n.jjtAddChild(a, 0);
		return n;
	}

	public static CodeNode slist(CodeNode s,CodeNode b)
	{
		if(s != null)
		{
			if(s instanceof CNSList)
			{
				if(b != null)
					s.jjtAddChild(b, s.jjtGetNumChildren());
				return s;
			}
			else
			{
				CodeNode n = new CNSList(code.parser.CodeParserTreeConstants.JJTSLIST);
				n.jjtAddChild(s, 0);
				if(b != null)
					n.jjtAddChild(b, 1);
				return n;
			}
		}
		else
			return b;
	}

	public static CodeNode actions_x1(String x)
	{
		CNAsgn n = new CNAsgn(code.parser.CodeParserTreeConstants.JJTASGN);
		CNId id = new CNId(code.parser.CodeParserTreeConstants.JJTID);
		id.setName(x);
		CNInt i = new CNInt(code.parser.CodeParserTreeConstants.JJTINT);
		i.setValue(1);
		n.jjtAddChild(id, 0);
		n.jjtAddChild(i, 1);
		return n;
	}

	public static CodeNode actions_xy_1(String x, String y)
	{
		CNAsgn n = new CNAsgn(code.parser.CodeParserTreeConstants.JJTASGN);
		CNId id = new CNId(code.parser.CodeParserTreeConstants.JJTID);
		id.setName(x);
		CNSub s = new CNSub(code.parser.CodeParserTreeConstants.JJTSUB);
		CNInt i = new CNInt(code.parser.CodeParserTreeConstants.JJTINT);
		CNId d = new CNId(code.parser.CodeParserTreeConstants.JJTID);
		d.setName(y);
		i.setValue(1);
		s.jjtAddChild(d, 0);
		s.jjtAddChild(i, 1);
		n.jjtAddChild(id, 0);
		n.jjtAddChild(s, 1);
		return n;
	}

	public static CodeNode actions_xy(String x, String y)
	{
		CNAsgn n = new CNAsgn(code.parser.CodeParserTreeConstants.JJTASGN);
		CNId id = new CNId(code.parser.CodeParserTreeConstants.JJTID);
		id.setName(x);
		CNId i = new CNId(code.parser.CodeParserTreeConstants.JJTID);
		i.setName(y);
		n.jjtAddChild(id, 0);
		n.jjtAddChild(i, 1);
		return n;
	}

	public static CodeNode actions_xxy(String x, String y)
	{
		CNAsgn n = new CNAsgn(code.parser.CodeParserTreeConstants.JJTASGN);
		CNId id = new CNId(code.parser.CodeParserTreeConstants.JJTID);
		id.setName(x);
		CNAdd s = new CNAdd(code.parser.CodeParserTreeConstants.JJTADD);
		CNId i = new CNId(code.parser.CodeParserTreeConstants.JJTID);
		i.setName(y);
		s.jjtAddChild(id, 0);
		s.jjtAddChild(i, 1);
		n.jjtAddChild(id, 0);
		n.jjtAddChild(s, 1);
		return n;
	}


	public static CodeNode guard_x1(String x)
	{
		CNEq n = new CNEq(code.parser.CodeParserTreeConstants.JJTEQ);
		CNId id = new CNId(code.parser.CodeParserTreeConstants.JJTID);
		id.setName(x);
		CNInt i = new CNInt(code.parser.CodeParserTreeConstants.JJTINT);
		i.setValue(1);
		n.jjtAddChild(id, 0);
		n.jjtAddChild(i, 1);
		return n;
	}

	public static CodeNode guard_x(String x)
	{
		CNGt n = new CNGt(code.parser.CodeParserTreeConstants.JJTGT);
		CNId id = new CNId(code.parser.CodeParserTreeConstants.JJTID);
		id.setName(x);
		CNInt i = new CNInt(code.parser.CodeParserTreeConstants.JJTINT);
		i.setValue(1);
		n.jjtAddChild(id, 0);
		n.jjtAddChild(i, 1);
		return n;
	}

}
